#include <stdio.h> 
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h> //for having fun with PIDs
#include <sys/wait.h>


#define DEBUG			0		//1 means: print out debug informations
#define	MAX_TOKENS		100		//maximum of words, entered in the command-line
#define MAX_SUBFOLDERS	100		//maximum of folders, that the current working directory can have
#define	MAX_WAIT_PIDS	100		//maximum of pids, the user can wait for (by entering "wait pid1 pid2 ..pidMAX_WAIT_PIDS")

int user_trys_to_abort = 0;

/******************************* strtoken ************************************
*                                                                            *
*   String 'str' is going to be separated into tokens 'token'   	     *
*   by separator 'separator'. The function returns the count of tokens	     *
*                                                                            *
*****************************************************************************/
int strtoken(char *str, char *separator, char *token[])
{
	int i = 0;
	
	token[i] = strtok(str, separator);
	
	while ( token[i] ) {
		i++;
		token[i] = strtok(NULL, separator);
	}
	return ( i );
}


/******************************* print_token *********************************
*                                                                            *
*   Print token is for testing purpose only.                                 *
*   It prints out the tokens stored in the given array.                      *
*                                                                            *
*                                                                            *
*****************************************************************************/
void print_token(int num_char, char *token[]) {
    int i = 0; 
	while ( i < num_char) {
		printf("\tPrintToken %i: '%s'\n", i, token[i]);
		i++;
	}
}


/******************************* get_current_dir *****************************
*                                                                            *
*  get the current directory                                                 *
*                                                                            *
*                                                                            *
*****************************************************************************/ 
void get_current_dir(char target[]) {
	size_t current_dir_length = 0;
	char* current_dir_buffer;
	char* current_dir;
	
	/*
	 * Read the current working directory
	 */
	current_dir_length = pathconf(".", _PC_PATH_MAX);
	if ((current_dir_buffer = (char*)malloc(current_dir_length)) != NULL) {
		current_dir = getcwd(current_dir_buffer, current_dir_length);
		free(current_dir_buffer);
		
		//copy the string into destination
		memcpy(target, current_dir, strlen(current_dir)+1);
	}
}


/******************************* display_name_and_dir ************************
*                                                                            *
*  this function displays the name and the current working directory         *
*  (as the prompt)                                                           *
*                                                                            *
*****************************************************************************/ 
void display_name_and_dir(char root_directory[]) {
	
	/*
	 * get the username
	 */
	char* user_name = getlogin();
	
	/*
	 * get the current directory
	 */
	char current_directory[FILENAME_MAX];
	get_current_dir(current_directory);
	
	/*
	 * get the current directory as tokens
	 */
	char* current_directory_tokens[MAX_SUBFOLDERS];
	int current_count = strtoken(current_directory, "/", current_directory_tokens);

	/*
	 * mem-copy the root-directory
	 */
	char origin[FILENAME_MAX];
	memcpy(origin, root_directory, strlen(root_directory)+1);
	
	/*
	 * get the root directory as tokens
	 */
	char* root_directory_tokens[MAX_SUBFOLDERS];
	int root_count = strtoken(origin, "/", root_directory_tokens);
	
	/*
	 * calculate the relative directory
	 */
	char relative_directory[FILENAME_MAX] = "";
	
	
	int k = 0;
	int m = 0;
	while ((k<=current_count) && (k<=root_count)) {	//for every token (foldername)

		//the root_directory is "/" ... we can use the absolute path (and doing a "." before)
		if (root_count == 0) {
			//begin with "."
			strcat(relative_directory, ".");
			
			//copy the absolute path
			get_current_dir(current_directory);
			strcat(relative_directory, current_directory);
			
			//computing the relative path is now completed... break out of the loop
			break;			
		} 
		
		//we are currently in the directory "/"
		if (current_count == 0) {
			//how many "../"es
			m = 0;
			while (m < root_count) {
				#if DEBUG
					printf("\tTHIS IS A HACK AT current_count=0 ;\tm=%i ; root_count=%i\n", m, root_count);
				#endif
				strcat(relative_directory, "../");
				m++;
				if (m == root_count-1) {	//THIS IS A HACK... I DON'T KNOW WHY root_count will be increased, when m = root_count
					strcat(relative_directory, "../");
					break;
				}
			}
			
			//computing the relative path is now completed... break out of the loop
			break;
		}
				
		//if the k's token (foldername) is not equal, then the relative path can be constructed
		if (strcmp(root_directory_tokens[k], current_directory_tokens[k]) != 0) {
			
			//how many paths '..'?
			m = k;
			while (m < root_count) {
				#if DEBUG
					printf("\tTHIS IS A HACK AT m=root_count-1; \tm=%i ; root_count=%i\n", m, root_count);
				#endif
				strcat(relative_directory, "../");
				m++;
				if (m == root_count-1) {	//THIS IS A HACK... I DON'T KNOW WHY root_count will be increased, when m = root_count
					strcat(relative_directory, "../");
					break;
				}
			}
			
			//write the rest of the current dir after the "../"es
			m = k;
			while (m < current_count) {
				strcat(relative_directory, current_directory_tokens[m]);
				strcat(relative_directory, "/");
				m++;
			}
			
			//computing the relative path is now completed... break out of the loop
			break;
		}
		
		//if we changed dir down from the root (then k is the last [root_count-1])
		if (k == root_count-1) {

			//begin with "./"
			strcat(relative_directory, "./");

			//write the rest of the current dir after the "./"
			m = k+1;
			while (m < current_count) {
				strcat(relative_directory, current_directory_tokens[m]);
				strcat(relative_directory, "/");
				m++;
			}
			
			//computing the relative path is now completed... break out of the loop
			break;
		}
		
		//if we changed dir up from the root and there is no "change into subfolder" 
		//(then k is the last [current_count-1])
		if (k == current_count-1) {
			
			//how many paths '..'?
			m = k+1;
			while (m < root_count) {
				strcat(relative_directory, "../");
				m++;
			}
						
			//computing the relative path is now completed... break out of the loop
			break;
		}
		
		k++;	//increase k (go to the next folder and compare their names)
	}
	
	/*
	 * display the username and current working directory (relative to the root-directory)
	 */
	printf("\033[22;32m%s", user_name);
	printf("\033[0m@");
	printf("\033[22;31m%s", relative_directory);
	printf("\033[0m > ");
}


/******************************* find_pipe ***********************************
*                                                                            *
*   find pipe searches a given char array for the '|'-character.             *
*   If '|' is found, it returns the position at which it is found.           *
*   If '|' isn't found, it returns -1.                                       *
*                                                                            *
*****************************************************************************/
int find_pipe(int num_token, char *token[]) {;
	int i = 0;
	while (i < num_token) {
		if( strcmp(token[i], "|") == 0 ) {
			return i;
		}
		i++;
	}
	
	return -1;
}

/******************************* ctrl_c_callback *****************************
*                                                                            *
* this function will be called, when the user pressed "ctrl-c"               *
*                                                                            *
*****************************************************************************/
void ctrl_c_callback(int sig) {
	#if DEBUG
		printf("\nctrl-c typed\n\n");	//for testing purposes only
	#endif
	
	user_trys_to_abort = 1;	//notice, that the user want's to abort
	
	signal(SIGINT, ctrl_c_callback);	//use this callback-function again
}


/******************************* is_last_token_and  **************************
*                                                                            *
*   is_last_token_and returns 1 if the last token is '&'                     *
*   else, it returns 0.                                                      *
*                                                                            *
*                                                                            *
*****************************************************************************/
int is_last_token_and(int num_char, char *token[]) {
    if ( strcmp(token[num_char-1], "&") == 0 ) {
		return 1;
	}
	return 0;
}


/******************************* print_process_infos *************************
*                                                                            *
*   this function prints information about a (terminated) child process      *
*                                                                            *
*                                                                            *
*****************************************************************************/
void print_process_infos(int child_pid, int status) {
	printf("\n\t[%d] TERMINATED\n", child_pid);
	printf("\t-----------------------------\n");
	printf("\t[%d] EXIT STATUS: %d\n", child_pid, WEXITSTATUS(status));
	printf("\t[%d] SIGNALED EXIT: %d\n", child_pid, WIFSIGNALED(status));
	if (WIFSIGNALED(status)) {
		printf("\t\t[%d] --> SIGNAL NUMBER, THAT CAUSED THE EXIT: %d\n", child_pid, WTERMSIG(status));
	}
	printf("\t[%d] PRODUCED A CORE DUMP: %d\n", child_pid, WCOREDUMP(status));
	printf("\t[%d] CHILD PROCESS STOPPED: %d\n", child_pid, WIFSTOPPED(status));
	if (WIFSTOPPED(status)) {
		printf("\t\t[%d] --> SIGNAL NUMBER, THAT CAUSED THE STOP: %d\n", child_pid, WSTOPSIG(status));
	}
	printf("\n");
}


/******************************* start_fork  **********************************
*				                                              *
*   this function forks the program, and waits (if necessary) for	      *
*   for the return							      *
*                                                                             *
******************************************************************************/
pid_t start_fork(char* token[], int num_token, int wait_for_process) {
	
	//starting the program
	#if DEBUG
		printf("doing the fork\n");
	#endif
	
	pid_t pid = vfork();
	if (pid < 0) {
		exit(-2);	//Error while forking program
	}
	
	if (pid == 0) {
		// I'm the child
		int ret = 0;
		
		if (wait_for_process) {
			#if DEBUG
				printf("Child pid: %d\n-->execvp'ing \"%s\" with %i arguments\n", getpid(), token[0], (num_token-1));
			#endif
			
			ret = execvp(token[0], token);	//execute the command (with all arguments)
			
		} else {	//if the function shall not wait, then the last token is an ampercent ... we have to delete it
			#if DEBUG
				printf("Child pid: %d\nnow collecting arguments...\n",getpid());
			#endif
			
			char* arguments[MAX_TOKENS];
			int arg = 0;
			while (arg < (num_token-1)) {
				//memcpy(arguments[arg], token[arg], strlen(token[arg]+1));	//WHY DOES THAT NOT WORK!?!?
				arguments[arg] = token[arg];	//must copy the reference to the string
				//printf("ARG %d: %s\n", arg, arguments[arg]);
				arg++;
			}
			
			#if DEBUG
				printf("...done collecting\n-->execvp'ing \"%s\" with %i arguments\n", token[0], (num_token-2));
			#endif
			
			signal(SIGINT, SIG_IGN);	//the child should not react on a ctrl-c signal
			ret = execvp(arguments[0], arguments);	//execute the command (with all arguments, but not the ampercent)
		}
		
		#if DEBUG
			printf("[%d] Child process exiting with code: %d", getpid(), ret);
		#endif
		
		exit(ret);	//end the child
		
	} else {
		//I'm NOT the child
		if (wait_for_process) {	//is it requested to wait for the process (no ampercent at the end)
			#if DEBUG
				printf("waiting for the pid: [%d]\n", pid);
			#endif
			
			int status;
			waitpid(pid, &status, 0);
			
			#if DEBUG
				printf("[%d] - Child has returned, status: %i\n", pid, status);
			#endif
			
		}
		
		return pid;
	}
}


/******************************* processing_pipe ******************************
 *                                                                            *
 * this function processes the "pipe" request of the user		      *
 *                                                                            *
 *****************************************************************************/
void processing_pipe(int num_token, char* token[], int pipe_position) {
	int fd[2];
	
	//create a new array with the left-hand exec-parameters (all in front the pipe_position)
	int cur_left = 0;
	char* arguments_left[pipe_position];
	while (cur_left < (num_token - pipe_position - 1)) {
		arguments_left[cur_left] = token[cur_left];	//must copy the reference to the string
		cur_left++;
	}
	arguments_left[cur_left] = NULL;
	
	//create a new array with the right-hand exec-parameters (all in front the pipe_position)
	int cur_right = pipe_position+1;
	char* arguments_right[num_token - pipe_position - 1];
	while (cur_right < num_token) {
		arguments_right[(cur_right - (pipe_position+1))] = token[cur_right];	//must copy the reference to the string
		//memcpy(arguments_right[(cur_right - (pipe_position+1))], token[cur_right], strlen(token[cur_right])+1);
		//printf("cur_right=%d: (%d).%s\n", cur_right, (cur_right - (pipe_position+1)), arguments_right[(cur_right - (pipe_position+1))]);
		cur_right++;
	}
	arguments_right[(num_token - pipe_position - 1)] = NULL;
	
	
	if (pipe(fd)) {	//open a pipe
		exit(3);	//if an error occurs exit this child
	}
	
	int pid = fork();	//create another child (because we have to pipe between 2 childs of the scurShell)
	if (pid < 0) {
		exit(-2);	//Error while forking program
	}
	
	if (pid == 0) {
		//I'm the child
		
		dup2(fd[0], 0);	//take the input out of the file-descriptor (pipe)
		close(fd[1]);	//clean (because it is unused)
		
		
		//char *grep_args[] = {arguments_right[0], arguments_right[1], NULL};
		int ret = execvp(arguments_right[0], arguments_right);	//execute the right-hand command
		perror(arguments_right[0]);	//display possible errors
		exit(ret);	//exit the child
		
		
	} else {
		//I'm NOT the child (but a child of the root-scurShell ... see the call in the main-method)
		
		dup2(fd[1], 1);	//use the pipe as stdout
		close(fd[0]);	//clean (because it is unused)
		
		int ret = execvp(arguments_left[0], arguments_left);	//execute the right-hand command
		perror(arguments_left[0]);	//display possible errors
		exit(ret);	//exit the child (.. i mean parent ... i mean child [of the root-scurShell])
	}
}


/******************************* main ****************************************
*                                                                            *
*   Here is where all the action is started.                                 *
*                                                                            *
*                                                                            *
*                                                                            *
*****************************************************************************/
int main(void) {
	char* token[MAX_TOKENS];	//here, the Token from the input line are stored 
	size_t user_line_length = 0;
	ssize_t user_line_size;
	char* user_line;	//the input line
	char root_directory[FILENAME_MAX];	//the shell was started in this directory
	
	get_current_dir(root_directory);	//saving the root directory for the first time

	printf("\033[22;32m\nShell\n");
	printf("---------\n");
		
	while(1) {
		//Step -2: flush the output (if there is some left)
		fflush(stdout);
		
		// Step -1: display the prompt
		display_name_and_dir(root_directory);
		
																					// Step 0: Get the userline and remove \n at the end (if possible)
		user_line_size = getline(&user_line, &user_line_length, stdin);
		if (user_line_size > 1) {	//it is possible to remove the \n
			user_line[strlen(user_line)-1] = '\0';
		}
		
																					// Step 1: Split input string into token
		int num_token = strtoken(user_line, " ",token); 
		#if DEBUG
			printf("num_token = %i;\ttoken[0] = '%s'\n", num_token, token[0]);
			printf("Step 1: Split input done... num_token: %d\n", num_token);
			printf("Step 1a: Printing tokens:\n");
			print_token(num_token, token);
		#endif
		
		int pos_pipe = -1;	//initialize the possible pipe-position
																					// Step 2: Look for the pipe(c)
		pos_pipe = find_pipe(num_token, token);
		
		#if DEBUG
			printf("Step 2: Searched for a pipe\n");
			printf("\tPipe position: %i (\"-1\" means 'no pipe')\n",pos_pipe);
		#endif
		
		if (pos_pipe != -1) {														// Step 3: Processing the pipe
			#if DEBUG
				printf("Step 3: Processing the pipe\n");
			#endif
			
			int pipe_pid = fork();
			if (pipe_pid < 0) {
				printf("error, while creating a child process for the pipe\n");
			} else if (pipe_pid == 0) {	//it's the child
				processing_pipe(num_token, token, pos_pipe);
			} else {	//it's the parent
				#if DEBUG
					printf("waiting for end of the pipe-child-process [%d]...\n", pipe_pid);
				#endif
				int pipe_status;
				waitpid(pipe_pid, &pipe_status, 0);
				usleep(1000*50);	//sleep a bit, for 
				#if DEBUG
					printf("...done waiting -> status is: %d\n", pipe_status);
				#endif
			}
			
			
		} else if ((num_token>0) && (is_last_token_and(num_token, token))) {		// Step 4: Is last Token "&" ???
			//if there is am ampercent at the end
			int new_child_pid = start_fork(token, num_token, 0);
			printf("the new CHILD PID is: %i\n", new_child_pid);	//output is required
			
			
		} else if ((num_token>0) && (strcmp(token[0], "exit")) == 0) {				// Step 5: Test if input is 'exit'
			
			printf("[%d] shell is going down\n", getpid());
			break;
			
		} else if ((num_token>0) && (strcmp(token[0], "cd")) == 0) {				// Step 6: Test if input is 'cd'
			if (num_token == 1) {	//without arguments
				#if DEBUG
					printf("changing directory to: %s (is the root-directory)\n", root_directory);
				#endif
				if (chdir(root_directory) == -1) {
					printf("\t'cd' was not possible\n");
				}
			} else {	//with one or more arguments
				#if DEBUG
					printf("changing directory to: %s\n", token[1]);
				#endif
				if (chdir(token[1]) == -1) {	//only the first argument is the valid one
					printf("\t'cd' was not possible\n");
				}
			}
		} else if ((num_token>0) && (strcmp(token[0], "wait") == 0)) {				// Step 7: Test if input is 'wait'

			pid_t required_pids[MAX_WAIT_PIDS];
			signal(SIGINT, ctrl_c_callback);	//enter to a callback function, when ctr-c (SIGINT) is pressed
			
			if (num_token < 2) {
				printf("YOU NEED TO ENTER A PID, TO WAIT FOR... SYNTAX: 'wait <pid1> <pid2>... <pidN>'\n");
			}
			
			//collect all pid's
			int pid_num = 1;	//init with 1, because 0 is the "wait" command
			while (pid_num < num_token) {
				required_pids[pid_num-1] = (pid_t)atoi(token[pid_num]);
				#if DEBUG
					printf("going to wait for pid %d: %d\n", pid_num, required_pids[pid_num-1]);			
				#endif
				pid_num++;
			}
			
			int status;
			int wStat;
			int current_pid;
			int displayed_pid_status = 0;
			while (1) {
				if (user_trys_to_abort) {
					break;	//stop waiting with breaking out of the loop
				}
				
				current_pid = 0;
				while (current_pid < pid_num) {	//check all pid's
					
					if (required_pids[current_pid] > 0) {
						wStat = waitpid(required_pids[current_pid], &status, WNOHANG|WUNTRACED);
				
						if (WIFEXITED(status) && (atoi(token[current_pid+1]) == wStat)) {	//if the called function exited
							print_process_infos(wStat, status);
							required_pids[current_pid] = (pid_t)(-1);	//do not display that pid's infos again
							displayed_pid_status++;
						}
				
						if (wStat < 0) {	//process was killed (by stop, or error, or interrupt, etc.)
							print_process_infos(atoi(token[current_pid+1]), status);
							required_pids[current_pid] = (pid_t)(-1);	//do not display that pid's infos again
							displayed_pid_status++;
						}
					}
					
					current_pid++;
				}
				
				if (displayed_pid_status == (pid_num-1)) {	//if all pid's status are already displayed
					break;
				}
				
				usleep(1000*50);	//sleep 0.05sec	(for relaxing the cpu)
			}
			
			//reset the defaults
			signal(SIGINT, SIG_DFL);	//default signal handling
			user_trys_to_abort = 0;		//reset the user-abortion-flag
			
		} else if (num_token>0) {												//Step 8: neither 'exit', nor...
			//'cd', nor 'wait', nor 'found &', nor 'found a pipe'
			
			printf("\n");
			
			if (!((num_token == 1) && (strcmp(token[0], "\n") == 0))) {
				//starting a fork with the given tokens
				start_fork(token, num_token, 1);
				//the last paramter "1" means, that this function is used as a blocking function
				//(the function call waits for the child-process to terminate)
			}
		}
	}

	return 0;
}
